"use strict";
var tslib_1 = require("tslib");
var ak_editor_prosemirror_1 = require("ak-editor-prosemirror");
var NodeType = (function (_super) {
    tslib_1.__extends(NodeType, _super);
    function NodeType() {
        return _super.apply(this, arguments) || this;
    }
    return NodeType;
}(ak_editor_prosemirror_1.NodeType));
function assert(expr, error) {
    if (!expr) {
        throw new Error(error);
    }
}
function isEmpty(properties) {
    var size = Array.isArray(properties)
        ? properties.length
        : Object.keys(properties).length;
    return size === 0;
}
function createMarkDef(schema, name) {
    var mark = schema.marks[name];
    var definition = {
        type: 'object',
        properties: {
            _: { 'enum': [name] },
        },
        required: ['_'],
        additionalProperties: false,
    };
    Object.keys(mark.attrs).forEach(function (key) {
        var attr = mark.attrs[key];
        definition['properties'][key] = { type: 'string' };
        if (attr.isRequired) {
            definition['required'].push(key);
        }
    });
    return definition;
}
var everHasMarks = function (schema, nodeName) { return Object
    .keys(schema.nodes)
    .map(function (nodeName) { return schema.nodes[nodeName]; })
    .map(function (node) { return node.contentExpr.elements; })
    .reduce(function (acc, x) { return acc.concat(x); }, [])
    .filter(function (contentElement) { return contentElement.nodeTypes
    .some(function (nodeType) { return nodeType.name === nodeName; }); })
    .map(function (contentElement) { return contentElement.marks; })
    .some(Boolean); };
function createTextNodeDef(schema, nodeName) {
    var definition = {
        properties: {
            text: {
                minLength: 1,
                type: 'string'
            },
            type: { 'enum': [nodeName] },
        },
        required: ['text', 'type'],
        additionalProperties: false,
    };
    if (everHasMarks(schema, nodeName)) {
        definition['properties']['marks'] = {};
    }
    return definition;
}
function createBlockNodeDef(schema, nodeName) {
    var _a = schema.nodes[nodeName].contentExpr, nodeType = _a.nodeType, elements = _a.elements;
    assert(elements.length <= 2, 'JSON Schema can not express more than two sequenced elements.');
    var definition = {
        properties: {
            type: { 'enum': [nodeName] },
        },
        required: ['type'],
        additionalProperties: false,
    };
    var makeNodeSchema = function (elem) {
        var nodeNames = elem.nodeTypes.map(function (type) { return type.name; });
        var refs = nodeNames.map(function (name) { return ({ '$ref': "#/definitions/" + name + "_node" }); });
        var nodes = refs.length === 1 ? refs[0] : { anyOf: refs };
        var markNames = elem.marks === true
            ? Object.keys(schema.marks)
            : (elem.marks || []).map(function (mark) { return mark.name; });
        if (markNames.length) {
            return {
                allOf: [
                    nodes,
                    {
                        properties: {
                            marks: {
                                type: 'array',
                                items: {
                                    anyOf: markNames.map(function (name) { return ({ $ref: "#/definitions/" + name + "_mark" }); }),
                                }
                            }
                        }
                    }
                ]
            };
        }
        else if (nodeNames.some(function (name) { return everHasMarks(schema, name); })) {
            return {
                allOf: [
                    nodes,
                    {
                        properties: {
                            marks: {
                                maxItems: 0
                            }
                        }
                    }
                ]
            };
        }
        else {
            return nodes;
        }
    };
    var content = {
        type: 'array',
        items: [],
        minItems: 0,
    };
    var first = elements[0], second = elements[1];
    if (first) {
        content['items'] = makeNodeSchema(first);
        content['minItems'] = first.min;
        if (second) {
            // It's not possible to express 'p{2} inline*'
            assert(content['minItems'] === 1, 'JSON Schema is unable to express an array with multiple specific leading items.');
            // Switch to tuple mode.
            content['items'] = [content['items']];
            content['additionalItems'] = makeNodeSchema(second);
            content['minItems'] += second.min;
        }
    }
    if (content['minItems'] === 0) {
        delete content['minItems'];
    }
    else {
        definition['required'].push('content');
    }
    if (!isEmpty(content['items'])) {
        definition['properties'].content = content;
    }
    definition['properties']['attrs'] = {
        type: 'object',
        properties: {},
        required: [],
    };
    Object.keys(nodeType.attrs).forEach(function (key) {
        var attr = nodeType.attrs[key];
        definition['properties']['attrs']['properties'][key] = {};
        if (attr.isRequired) {
            definition['properties']['attrs']['required'].push(key);
        }
    });
    if (definition['properties']['attrs']['required'].length > 0) {
        definition['required'].push('attrs');
    }
    else {
        delete definition['properties']['attrs']['required'];
    }
    if (isEmpty(definition['properties']['attrs']['properties'])) {
        delete definition['properties']['attrs'];
    }
    return definition;
}
function createNodeDef(schema, name) {
    var nodeType = schema.nodes[name].contentExpr.nodeType;
    if (nodeType instanceof ak_editor_prosemirror_1.Text) {
        return createTextNodeDef(schema, name);
    }
    else {
        return createBlockNodeDef(schema, name);
    }
}
function default_1(schema) {
    var nodeDefs = Object.keys(schema.nodes)
        .reduce(function (prev, name) {
        return (tslib_1.__assign({}, prev, (_a = {}, _a[name + "_node"] = createNodeDef(schema, name), _a)));
        var _a;
    }, {});
    var markDefs = Object.keys(schema.marks)
        .reduce(function (prev, name) {
        return (tslib_1.__assign({}, prev, (_a = {}, _a[name + "_mark"] = createMarkDef(schema, name), _a)));
        var _a;
    }, {});
    return {
        '$schema': 'http://json-schema.org/draft-04/schema#',
        'description': 'Schema for Atlassian Editor documents.',
        '$ref': '#/definitions/doc_node',
        'definitions': tslib_1.__assign({}, nodeDefs, markDefs)
    };
}
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = default_1;
//# sourceMappingURL=json-schema.js.map